/* ------------------------------------------------------------------------- */
/*
 *  timefmt.h
 *
 *  Copyright (c) 2004 - 2006 "clown", All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Thu 08 Jun 2006 14:18:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_TIMEFMT_H
#define CLX_TIMEFMT_H

#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include "lexical_cast.h"
#include "tokenmap.h"
#include "time.h"

#define CLX_TIMEFMT "%y/%m/%d %H:%M:%S"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_timefmt
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_timefmt {
	public:
		typedef time time_type;
		typedef time::value_type value_type;
		typedef CharT char_type;
		typedef typename std::basic_string<CharT, Traits> string_type;
		typedef typename string_type::const_iterator const_iterator;
		
		// constructor and destructor
		explicit basic_timefmt(const string_type& fmt = CLX_TIMEFMT) :
			t_(), tmap_(), fmt_(fmt), idset_() { this->init(); }
		
		explicit basic_timefmt(const time_type& t, const string_type& fmt = CLX_TIMEFMT) :
			t_(t), tmap_(), fmt_(fmt), idset_() { this->init(); }
		
		virtual ~basic_timefmt() {}
		
		// operator
		virtual void reset() { t_ = time_type(); }
		
		virtual time_type& assign(const string_type& s) {
			func f;
			f.assign(s, fmt_, idset_);
			
			for (typename string_type::const_iterator p = idset_.begin(); p != idset_.end(); p++) {
				if (f.find(*p) != f.end()) *tmap_[*p] = lexical_cast<value_type>(f[*p]);
			}
			return t_;
		}
		
		virtual time_type& assign(const string_type& s, const string_type& fmt) {
			fmt_ = fmt;
			return this->assign(s);
		}
		
		virtual string_type str() {
			string_type dest;
			for (const_iterator pf = fmt_.begin(); pf != fmt_.end(); pf++) {
				// if the last character is '%', treat this as same as "%%".
				if (*pf == rep && pf + 1 != fmt_.end() && *(pf + 1) != rep) {
					string_type flag;
					while (++pf != fmt_.end() && tmap_.find(*pf) == tmap_.end()) flag += *pf;
					if (pf == fmt_.end()) break; // uncorrect format
					dest += this->put(*tmap_[*pf], flag);
				}
				else dest += *pf;
			}
			return dest;
		}
		
		virtual string_type str(const string_type& fmt) {
			fmt_ = fmt;
			return this->str();
		}
		
		const time_type& data() const { return t_; }
		
	private:
		typedef basic_tokenmap<format_separator<CharT, Traits>, string_type,
			std::map<CharT, string_type> > func;
		typedef typename std::basic_stringstream<CharT, Traits> internal_stream;
		typedef typename std::map<CharT, value_type*> time_map;
		
		static const char_type rep = '%';
		static const char_type eos = '\0';
		static const char_type ws = ' ';
		static const char_type zero = '0';
		static const char_type lpos = '-';
		
		time_type t_;
		time_map tmap_;
		string_type fmt_;
		string_type idset_;
		
		void init() {
			tmap_['y'] = &t_.year;
			tmap_['m'] = &t_.mon;
			tmap_['d'] = &t_.day;
			tmap_['H'] = &t_.hour;
			tmap_['M'] = &t_.min;
			tmap_['S'] = &t_.sec;
			idset_ = "ymdHMS";
		}
		
		string_type put(value_type x, const string_type& flag) {
			internal_stream ss;
			
			setopt(ss, flag);
			ss << x;
			
			return ss.str();
		}
		
		void setopt(internal_stream& ss, const string_type& flag) {
			bool lp = false;
			for (const_iterator pos = flag.begin(); pos != flag.end(); pos++) {
				switch (*pos) {
				case lpos:
					lp = true;
					ss << std::setfill(ws);
					ss << std::left;
					break;
				case ws:
					ss << std::setfill(*pos);
					break;
				case zero:
					if (!lp) ss << std::setfill(*pos);
					break;
				default:
					if (isdigit(*pos)) {
						setwidth(ss, pos, flag.end());
						return;
					}
					break; /* ignore unknown flags */
				}
			}
		}
		
		void setwidth(internal_stream& ss, const_iterator& cur, const_iterator last) {
			std::string digit;
			for (; cur < last && isdigit(*cur); cur++) digit += *cur;
			if (!digit.empty()) ss << std::setw(lexical_cast<int>(digit));
		}
	};
	
	typedef basic_timefmt<char> timefmt;
}

#endif // CLX_TIMEFMT_H
