/* ------------------------------------------------------------------------- */
/*
 *  tokenizer_func.h
 *
 *  Copyright (c) 2004 - 2006 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Wed 07 Jun 2006 20:45:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_TOKENIZER_FUNC_H
#define CLX_TOKENIZER_FUNC_H

#include <algorithm>
#include <cstdlib>
#include <string>
#include "lexical_cast.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	/*!
	 *  char_separator
	 *
	 *  The char_separator class breaks a sequence of characters into
	 *  tokens based on the character delimiters.
	 */
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class char_separator {
	public:
		typedef CharT char_type;
		typedef typename std::basic_string<CharT, Traits> string_type;
		
		// constructor and destructor
		explicit char_separator(char_type c = ' ', bool x = false) :
			c_(1, c), emtok_(x) {}
		
		explicit char_separator(string_type c, bool x = false) :
			c_(c), emtok_(x) {}
		
		// operator
		void reset() { return; }
		
		template <class InputIterator, class Token>
		bool operator()(InputIterator& next, InputIterator last, Token& dest) {
			dest = Token();
			
			if (next == last) return 0;
			if (!emtok_) { // invalid empty token, so skip continuous delimiters.
				while (std::find(c_.begin(), c_.end(), *next) != c_.end()) {
					if (++next == last) return 0;
				}
			}
			
			while (next != last) { // obtain the next token.
				if (std::find(c_.begin(), c_.end(), *next) != c_.end()) break;
				dest += *next;
				next++;
			}
			return 1;
		}
		
	private:
		string_type c_;
		bool emtok_;
	
	};
	
	/* --------------------------------------------------------------------- */
	/*!
	 *  escape_separator
	 *
	 *  The escape_separator class is a super-set of what is commonly known
	 *  as a comma separated value (csv) list.
	 */
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class escape_separator {
	public:
		typedef CharT char_type;
		typedef typename std::basic_string<CharT, Traits> string_type;
		
		// constructor and destructor
		explicit escape_separator(char_type c = ',', char_type q = '\"',
			char_type e = '\\', bool x = false) :
			c_(1, c), quote_(1, q), esc_(1, e), emtok_(x) {}
		
		explicit escape_separator(string_type c, string_type q, string_type e, bool x = false) :
			c_(c), quote_(q), esc_(e), emtok_(x) {}
		
		// operator
		void reset() { return; }
		
		template <class InputIterator, class Token>
		bool operator()(InputIterator& next, InputIterator last, Token& dest) {
			bool inq = false;
			dest = Token();
			
			if (next == last) return 0;
			if (!emtok_) { // invalid empty token, so skip continuous delimiters.
				while (std::find(c_.begin(), c_.end(), *next) != c_.end()) {
					if (++next == last) return 0;
				}
			}
			
			while (next != last) { // obtain the next token.
				if (std::find(quote_.begin(), quote_.end(), *next) != quote_.end()) inq = !inq;
				else if (!inq && std::find(c_.begin(), c_.end(), *next) != c_.end()) break;
				else {
					if (std::find(esc_.begin(), esc_.end(), *next) != esc_.end()) next++;
					dest += *next;
				}
				next++;
			}
			return 1;
		}
		
	private:
		string_type c_;
		string_type quote_;
		string_type esc_;
		bool emtok_;
	};
	
	/* --------------------------------------------------------------------- */
	/*!
	 *  format_separator
	 *
	 *  The format separator class is split a string according to the
	 *  given format. The behavior of this class is similar to scanf()
	 *  function.
	 */
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class format_separator {
	public:
		typedef CharT char_type;
		typedef int int_type;
		typedef typename std::basic_string<CharT, Traits> string_type;
		typedef typename string_type::iterator iterator;
		
		// constructor and destructor
		explicit format_separator(const string_type& fmt = "%s", bool x = true) :
			fmt_(fmt), skipws_(x) {
			cur_ = fmt_.begin();
			ids_ = 's';
		}
		
		explicit format_separator(const string_type& fmt, const string_type& idset, bool x = true) :
			ids_(idset), fmt_(fmt), skipws_(x) { cur_ = fmt_.begin(); }
		
		virtual ~format_separator() {}
		
		// operator
		void reset() { cur_ = fmt_.begin(); }
		
		template <class InputIterator, class Token>
		int_type operator()(InputIterator& next, InputIterator last, Token& dest) {
			while (next != last && cur_ != fmt_.end()) {
				if (skipws_ && *next == ws && *cur_ != ws) next++;
				else if (*next == rep) {
					if (cur_ + 1 != fmt_.end() && *cur_ == rep && *(cur_ + 1) == rep) {
						next++;
						cur_ += 2;
					}
					else break;
				}
				else if (*cur_ == rep) {
					string_type flag;
					while (++cur_ != fmt_.end() && ids_.find(*cur_) == string_type::npos) flag += *cur_;
					if (cur_ == fmt_.end()) break;
					char_type key = *cur_++;
					char_type c = (cur_ != fmt_.end()) ? *cur_++ : eos;
					dest = this->get(next, last, c, flag);
					return static_cast<int_type>(key);
				}
				else if (*cur_ == *next) { next++; cur_++; }
				else break;
			}
			return 0;
		}
		
	private:
		static const char_type rep = '%';
		static const char_type eos = '\0';
		static const char_type ws = ' ';
		
		string_type ids_;
		string_type fmt_;
		iterator cur_;
		bool skipws_;
		
		template <class InputIterator>
		string_type get(InputIterator& next, InputIterator last, char_type c, const string_type& flag) {
			int n = -1;
			if (!flag.empty() && isdigit(flag.at(0))) n = lexical_cast<int>(flag);
			
			string_type dest;
			while (next != last && *next != c && n-- != 0) {
				dest += *next;
				next++;
			}
			return dest;
		}
	};
}

#endif // CLX_TOKENIZER_FUNC_H
