/* ------------------------------------------------------------------------- */
/*
 *  tokenmap.h
 *
 *  Copyright (c) 2004 - 2006 "clown", All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Thu 08 Jun 2006 11:44:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_TOKENMAP_H
#define CLX_TOKENMAP_H

#include <string>
#include <map>
#include "tokenizer_func.h"
#include "lexical_cast.h"

#define CLX_ALPHABET "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_tokenmap
	/* --------------------------------------------------------------------- */
	template <
		class FormatSeparator = format_separator<char>,
		class String = std::string,
		class Container = std::map<char, std::string>
	>
	class basic_tokenmap {
	public:
		typedef String string_type;
		typedef FormatSeparator func;
		typedef Container container;
		typedef typename container::value_type value_type;
		typedef typename container::key_type key_type;
		typedef typename container::mapped_type mapped_type;
		typedef typename container::size_type size_type;
		typedef typename container::const_iterator const_iterator;
		typedef typename container::const_reverse_iterator const_reverse_iterator;
		
		// constructor and destructor
		basic_tokenmap() :
			f_(), v_() {}
		
		explicit basic_tokenmap(const func& f) :
			f_(f), v_() {}
		
		explicit basic_tokenmap(const string_type& fmt, const string_type& idset = CLX_ALPHABET) :
			f_(fmt, idset), v_() {}
		
		virtual ~basic_tokenmap() {}
		
		//operator
		virtual void reset() {
			f_.reset();
			v_.clear();
		}
		
		virtual container& assign(const string_type& s) {
			typedef typename string_type::iterator string_iterator;
			
			string_type token;
			string_type tmp(s);
			string_iterator pos = tmp.begin();
			
			this->reset();
			key_type c;
			while (c = f_(pos, tmp.end(), token)) {
				v_[c] = lexical_cast<mapped_type>(token);
				if (pos++ == tmp.end()) break;
			}
			return v_;
		}
		
		virtual container& assign(const string_type& s, const func& f) {
			f_ = f;
			return this->assign(s);
		}
		
		virtual container& assign(const string_type& s, const string_type& fmt, const string_type& idset = CLX_ALPHABET) {
			func f(fmt, idset);
			f_ = f;
			return this->assign(s);
		}
		
		virtual container& operator()(const string_type& s) { return this->assign(s); }
		virtual container& operator()(const string_type& s, const func& f) { return this->assign(s, f); }
		virtual container& operator()(const string_type& s, const string_type& fmt, const string_type& idset = CLX_ALPHABET) { return this->assign(s, fmt, idset); }
		
		bool empty() const { return v_.empty(); }
		size_type size() const { return v_.size(); }
		
		// access each token.
		mapped_type operator[](const key_type& x) { return v_[x]; }
		const_iterator begin() const { return v_.begin(); }
		const_iterator end() const { return v_.end(); }
		const_reverse_iterator rbegin() const { return v_.rbegin(); }
		const_reverse_iterator rend() const { return v_.rend(); }
		const_iterator find(const key_type& x) const { return v_.find(x); }
		
	private:
		func f_;
		container v_;
	};
	
	typedef basic_tokenmap<> tokenmap;
}

#endif // CLX_TOKENMAP_H
